import pytest
from django.conf import settings

from sage_invoice.check import (
    check_installed_apps,
    check_required_libraries,
    check_required_settings,
)


@pytest.mark.django_db
class TestSystemChecks:

    @pytest.fixture
    def reset_settings(self):
        """Fixture to reset settings after each test."""
        original_settings = settings.INSTALLED_APPS[:]
        yield
        settings.INSTALLED_APPS = original_settings

    def test_check_installed_apps_success(self, reset_settings):
        settings.INSTALLED_APPS = list(settings.INSTALLED_APPS) + ["sage_invoice"]
        errors = check_installed_apps(None)
        assert errors == []

    def test_check_installed_apps_failure(self, reset_settings):
        settings.INSTALLED_APPS = []
        errors = check_installed_apps(None)
        assert len(errors) == 1
        assert errors[0].id == "sage_invoice.E001"
        assert errors[0].msg == "'sage_invoice' is missing in INSTALLED_APPS."

    @pytest.fixture
    def mock_find_spec(self, monkeypatch):
        """Fixture to mock find_spec."""

        def mock_spec(name):
            if name == "sage_invoice":
                return True
            return None

        monkeypatch.setattr("importlib.util.find_spec", mock_spec)

    def test_check_required_libraries_success(self, mock_find_spec):
        errors = check_required_libraries(None)
        assert errors == []

    def test_check_required_libraries_failure(self, monkeypatch):

        def mock_spec(name):
            return None  # Simulate library not being installed

        monkeypatch.setattr("importlib.util.find_spec", mock_spec)
        errors = check_required_libraries(None)
        assert len(errors) == 0

    def test_check_required_settings_success(self):
        settings.SAGE_MODEL_PREFIX = "invoice"
        settings.SAGE_MODEL_TEMPLATE = "sage_invoice"
        errors = check_required_settings(None)
        assert errors == []

    def test_check_required_settings_failure(self):
        if hasattr(settings, "SAGE_MODEL_PREFIX"):
            del settings.SAGE_MODEL_PREFIX
        if hasattr(settings, "SAGE_MODEL_TEMPLATE"):
            del settings.SAGE_MODEL_TEMPLATE

        errors = check_required_settings(None)
        assert len(errors) == 2
        assert errors[0].id == "sage_invoice.E003"
        assert (
            errors[0].msg == "The required setting 'SAGE_MODEL_PREFIX' is not defined."
        )
        assert errors[1].id == "sage_invoice.E004"
        assert (
            errors[1].msg
            == "The required setting 'SAGE_MODEL_TEMPLATE' is not defined."
        )



# serializers.py
from rest_framework import serializers

from sage_invoice.models import Item


class ItemSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Item
        fields = "__all__"
        lookup_field = "id"
        extra_kwargs = {
            "url": {"lookup_field": "id"},
            "invoice": {"lookup_field": "slug"},
        }



from django.db import models


class InvoiceStatus(models.TextChoices):
    PAID = ("paid", "PAID")
    UNPAID = ("unpaid", "UNPAID")


class Currency(models.TextChoices):
    USD = ("USD", "US Dollar")
    EUR = ("EUR", "Euro")
    GBP = ("GBP", "British Pound")
    JPY = ("JPY", "Japanese Yen")
    AUD = ("AUD", "Australian Dollar")
    CAD = ("CAD", "Canadian Dollar")
    CHF = ("CHF", "Swiss Franc")
    CNY = ("CNY", "Chinese Yuan")
    INR = ("INR", "Indian Rupee")
    RUB = ("RUB", "Russian Ruble")
    AED = ("AED", "UAE Dirham")
    SAR = ("SAR", "Saudi Riyal")
    TRY = ("TRY", "Turkish Lira")
    BRL = ("BRL", "Brazilian Real")
    ZAR = ("ZAR", "South African Rand")
    NZD = ("NZD", "New Zealand Dollar")
    KRW = ("KRW", "South Korean Won")
    SGD = ("SGD", "Singapore Dollar")
    MXN = ("MXN", "Mexican Peso")
    IRR = ("IRR", "Iranian Rial")
    TOMAN = ("TOMAN", "Iranian Toman")
    QAR = ("QAR", "Qatari Riyal")
    KWD = ("KWD", "Kuwaiti Dinar")
    BHD = ("BHD", "Bahraini Dinar")
    OMR = ("OMR", "Omani Rial")
    EGP = ("EGP", "Egyptian Pound")



from rest_framework.exceptions import NotFound
from rest_framework.versioning import BaseVersioning

SUPPORTED_VERSIONS = ["1.0"]


class HeaderVersioning(BaseVersioning):
    def determine_version(self, request, *args, **kwargs):
        # Get the version from the 'X-API-Version' header
        accept_header = request.META.get("HTTP_X_API_VERSION", "")

        # Check if the version header contains 'v'
        if "v" in accept_header:
            version = accept_header.split("v")[-1].split("+")[0]
        else:
            version = None

        # If no version is found, default to the latest version
        if not version:
            return SUPPORTED_VERSIONS[-1]  # Latest version

        # If version is not supported, raise an error
        if version not in SUPPORTED_VERSIONS:
            raise NotFound(f"API version '{version}' is not supported.")

        return version



import pytest
from unittest.mock import patch, MagicMock
from jinja2.exceptions import TemplateNotFound
from sage_invoice.service.invoice_create import QuotationService
from sage_invoice.models import Invoice, Item, Expense
from sage_invoice.service.total import ExpenseService
from unittest import mock
from decimal import Decimal
from django.core.exceptions import ValidationError


@pytest.mark.django_db
class TestQuotationService:

    @patch('sage_invoice.service.invoice_create.JinjaTemplateDiscovery')
    def test_init(self, mock_template_discovery):
        # Test if QuotationService initializes correctly
        service = QuotationService()
        assert service.template_discovery == mock_template_discovery.return_value
        assert service.env is not None
        mock_template_discovery.assert_called_once_with(
            sage_template_dir="sage_invoice"  # Default directory
        )

    @pytest.fixture
    def invoice(self, db):
        # Create a sample invoice and related items
        invoice = Invoice.objects.create(
            title="Test Invoice",
            invoice_date="2024-09-10",
            customer_name="Test Customer",
            tracking_code="INV-2024",
            status="unpaid",
            currency="USD",
            due_date="2024-10-10"
        )
        Expense.objects.create(
            invoice=invoice,
            subtotal=100,
            tax_percentage=10,
            discount_percentage=5,
            concession_percentage=0,
            tax_amount=10,
            discount_amount=5,
            concession_amount=0,
            total_amount=105
        )
        Item.objects.create(
            invoice=invoice,
            description="Test Item",
            quantity=1,
            unit_price=100,
            total_price=100
        )
        return invoice

    @patch('sage_invoice.service.invoice_create.JinjaTemplateDiscovery.get_template_path')
    def test_render_quotation_success(self, mock_get_template_path, invoice):
        # Mock template discovery and rendering process
        mock_template = MagicMock()
        mock_template.render.return_value = "Rendered HTML"
        mock_get_template_path.return_value = "path/to/template.html"

        service = QuotationService()
        service.env.get_template = MagicMock(return_value=mock_template)

        # Mock the queryset with first method
        mock_queryset = MagicMock()
        mock_queryset.first.return_value = invoice

        result = service.render_quotation(mock_queryset)

        assert result == "Rendered HTML"
        mock_get_template_path.assert_called_once()
        service.env.get_template.assert_called_once_with("template.html")
        mock_template.render.assert_called_once()

    @patch('sage_invoice.service.invoice_create.JinjaTemplateDiscovery.get_template_path')
    def test_render_quotation_template_not_found(self, mock_get_template_path, invoice):
        # Test case when template is not found
        mock_get_template_path.return_value = None
        service = QuotationService()

        # Mock the queryset with first method
        mock_queryset = MagicMock()
        mock_queryset.first.return_value = invoice

        with pytest.raises(TemplateNotFound):
            service.render_quotation(mock_queryset)

        mock_get_template_path.assert_called_once()

    def test_render_context(self, invoice):
        # Test context generation for the invoice
        service = QuotationService()

        # Ensure contacts and other fields are handled correctly
        invoice.contacts = ["you@example.com", "1234567890"]
        context = service.render_context(invoice)

        assert context["title"] == invoice.title
        assert context["tracking_code"] == invoice.tracking_code
        assert context["subtotal"] == invoice.total.subtotal
        assert context["grand_total"] == invoice.total.total_amount
        assert context["customer_name"] == invoice.customer_name
        assert len(context["items"]) == invoice.items.count()

    def test_invalid_invoice_due_date(self):
        # Test validation on due date
        invoice = Invoice(
            title="Invalid Invoice",
            invoice_date="2024-09-10",
            due_date="2024-09-05",
            customer_name="Test Customer",
            tracking_code="INV-2024",
            status="unpaid",
            currency="USD"
        )
        with pytest.raises(ValidationError):
            invoice.clean()

@pytest.mark.django_db
class TestExpenseService:

    @pytest.fixture
    def invoice(self, db):
        """Fixture to create a real Invoice instance with items."""
        invoice = Invoice.objects.create(
            title="Test Invoice",
            invoice_date="2024-08-25",
            customer_name="John Doe",
            status="unpaid",
            due_date="2024-09-01",
        )
        invoice.items.create(
            description="Item 1", quantity=1, unit_price=Decimal("100.00")
        )
        invoice.items.create(
            description="Item 2", quantity=2, unit_price=Decimal("50.00")
        )
        return invoice

    @pytest.fixture
    def invoice_total(self, invoice):
        """Fixture to create an Expense instance."""
        return Expense.objects.create(
            tax_percentage=Decimal("10.00"),
            discount_percentage=Decimal("5.00"),
            invoice=invoice,
        )

    def test_calculate_and_save(self, invoice_total):
        """Test the calculate_and_save method of ExpenseService."""
        service = ExpenseService()

        with mock.patch.object(Expense, "save") as mock_save:
            service.calculate_and_save(invoice_total)

            # Verify calculations
            assert invoice_total.subtotal == Decimal("200.00")
            assert invoice_total.tax_amount == Decimal("20.00")
            assert invoice_total.discount_amount == Decimal("10.00")
            assert invoice_total.total_amount == Decimal("210.00")
            mock_save.assert_called_once()

    def test_calculate_and_save_with_no_items(self, db):
        """Test calculate_and_save when the invoice has no items."""
        invoice = Invoice.objects.create(
            title="Empty Invoice",
            invoice_date="2024-08-25",
            customer_name="John Doe",
            status="unpaid",
            due_date="2024-09-01",
        )

        invoice_total = Expense.objects.create(
            tax_percentage=Decimal("10.00"),
            discount_percentage=Decimal("5.00"),
            invoice=invoice,
        )

        service = ExpenseService()

        with mock.patch.object(invoice_total, "save") as mock_save:
            service.calculate_and_save(invoice_total)

            # Verify calculations with no items
            assert invoice_total.subtotal == Decimal("0.00")
            assert invoice_total.tax_amount == Decimal("0.00")
            assert invoice_total.discount_amount == Decimal("0.00")
            assert invoice_total.total_amount == Decimal("0.00")
            mock_save.assert_called_once()



# Generated by Django 5.1.2 on 2024-10-17 17:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sage_invoice", "0002_alter_invoice_template_choice"),
    ]

    operations = [
        migrations.AlterField(
            model_name="expense",
            name="invoice",
            field=models.OneToOneField(
                db_comment="Reference to the associated invoice",
                help_text="The invoice associated with this total.",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="expense",
                to="sage_invoice.invoice",
                verbose_name="Invoice",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="category",
            field=models.ForeignKey(
                blank=True,
                db_comment="Category associated with this invoice",
                help_text="The category associated with this invoice.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="invoices",
                to="sage_invoice.category",
                verbose_name="Category",
            ),
        ),
        migrations.AlterField(
            model_name="item",
            name="quantity",
            field=models.PositiveIntegerField(
                blank=True,
                db_comment="The quantity of the invoice item",
                help_text="The quantity of the item.",
                null=True,
                verbose_name="Quantity",
            ),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class InvoiceConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "sage_invoice"
    verbose_name = _("Invoice Management")

    def ready(self) -> None:
        import sage_invoice.check
        import sage_invoice.signals



# Generated by Django 5.1 on 2024-10-08 14:44

import django.db.models.deletion
import django_jsonform.models.fields
from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        db_comment="Stores the unique title of the instance.",
                        help_text="Enter a unique title.",
                        max_length=255,
                        unique=True,
                        verbose_name="Title",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        allow_unicode=True,
                        db_comment="Stores the URL-friendly slug derived from the title.",
                        help_text="URL-friendly slug from the title.",
                        max_length=255,
                        unique=True,
                        verbose_name="Slug",
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        blank=True,
                        db_comment="Description of the Category",
                        help_text="Description of the Category.",
                        max_length=255,
                        null=True,
                        verbose_name="Description",
                    ),
                ),
            ],
            options={
                "verbose_name": "Category",
                "verbose_name_plural": "Categories ",
                "db_table": "sage_invoice_cat",
            },
        ),
        migrations.CreateModel(
            name="Invoice",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        db_comment="Stores the unique title of the instance.",
                        help_text="Enter a unique title.",
                        max_length=255,
                        unique=True,
                        verbose_name="Title",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        allow_unicode=True,
                        db_comment="Stores the URL-friendly slug derived from the title.",
                        help_text="URL-friendly slug from the title.",
                        max_length=255,
                        unique=True,
                        verbose_name="Slug",
                    ),
                ),
                (
                    "invoice_date",
                    models.DateField(
                        db_comment="Invoice date created",
                        help_text="The date when the invoice was created.",
                        verbose_name="Invoice Date",
                    ),
                ),
                (
                    "customer_name",
                    models.CharField(
                        db_comment="Customer name created",
                        help_text="The name of the customer.",
                        max_length=255,
                        verbose_name="Customer Name",
                    ),
                ),
                (
                    "tracking_code",
                    models.CharField(
                        db_comment="Tracking code created",
                        help_text="Enter the first 3-4 characters of the tracking code. The full code will be auto-generated as <prefix> + <date> + <random number>.",
                        max_length=255,
                        verbose_name="Tracking Code",
                    ),
                ),
                (
                    "contacts",
                    django_jsonform.models.fields.JSONField(
                        blank=True, null=True, verbose_name="Customer Contacts"
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[("paid", "PAID"), ("unpaid", "UNPAID")],
                        db_comment="Current status of the invoice (Paid, Unpaid)",
                        help_text="The current status of the invoice (Paid, Unpaid).",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                (
                    "receipt",
                    models.BooleanField(
                        db_comment="Check if this invoice is a receipt",
                        default=False,
                        help_text="Is this a receipt or an invoice",
                        verbose_name="Receipt",
                    ),
                ),
                (
                    "notes",
                    django_jsonform.models.fields.JSONField(
                        blank=True,
                        db_comment="This field stores additional dynamic content in JSON format. ",
                        help_text="\n            You can add any number of custom fields dynamically,\n            such as'Terms & Conditions', 'Technology Tips', etc.\n            ",
                        null=True,
                        verbose_name="Notes",
                    ),
                ),
                (
                    "currency",
                    models.CharField(
                        choices=[
                            ("USD", "US Dollar"),
                            ("EUR", "Euro"),
                            ("GBP", "British Pound"),
                            ("JPY", "Japanese Yen"),
                            ("AUD", "Australian Dollar"),
                            ("CAD", "Canadian Dollar"),
                            ("CHF", "Swiss Franc"),
                            ("CNY", "Chinese Yuan"),
                            ("INR", "Indian Rupee"),
                            ("RUB", "Russian Ruble"),
                            ("AED", "UAE Dirham"),
                            ("SAR", "Saudi Riyal"),
                            ("TRY", "Turkish Lira"),
                            ("BRL", "Brazilian Real"),
                            ("ZAR", "South African Rand"),
                            ("NZD", "New Zealand Dollar"),
                            ("KRW", "South Korean Won"),
                            ("SGD", "Singapore Dollar"),
                            ("MXN", "Mexican Peso"),
                            ("IRR", "Iranian Rial"),
                            ("TOMAN", "Iranian Toman"),
                            ("QAR", "Qatari Riyal"),
                            ("KWD", "Kuwaiti Dinar"),
                            ("BHD", "Bahraini Dinar"),
                            ("OMR", "Omani Rial"),
                            ("EGP", "Egyptian Pound"),
                        ],
                        db_comment="Which currency is this item ",
                        default="USD",
                        help_text="Currency of unit price",
                        max_length=10,
                        verbose_name="Currency",
                    ),
                ),
                (
                    "due_date",
                    models.DateField(
                        db_comment="Due date of the invoice",
                        help_text="The date by which the invoice should be paid.",
                        verbose_name="Due Date",
                    ),
                ),
                (
                    "logo",
                    models.ImageField(
                        blank=True,
                        db_comment="Logo displayed on the invoice",
                        help_text="The logo displayed on the invoice.",
                        null=True,
                        upload_to="invoices/logos/",
                        verbose_name="Logo",
                    ),
                ),
                (
                    "signature",
                    models.ImageField(
                        blank=True,
                        db_comment="Signature image for the invoice",
                        help_text="The signature image for the invoice.",
                        null=True,
                        upload_to="invoices/signatures/",
                        verbose_name="Signature",
                    ),
                ),
                (
                    "stamp",
                    models.ImageField(
                        blank=True,
                        db_comment="Stamp image for the invoice",
                        help_text="The stamp image for the invoice.",
                        null=True,
                        upload_to="invoices/stamps/",
                        verbose_name="Stamp",
                    ),
                ),
                (
                    "template_choice",
                    models.CharField(
                        choices=[
                            ("1", "Invoice1 Template"),
                            ("2", "Invoice2 Template"),
                            ("3", "Invoice3 Template"),
                            ("4", "Invoice4 Template"),
                        ],
                        db_comment="Template choice for the invoice",
                        help_text="The template you want for your invoice",
                        max_length=20,
                        verbose_name="Template choice",
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        blank=True,
                        db_comment="Category associated with this invoice",
                        help_text="The category associated with this invoice.",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="category",
                        to="sage_invoice.category",
                        verbose_name="Category",
                    ),
                ),
            ],
            options={
                "verbose_name": "Invoice",
                "verbose_name_plural": "Invoices",
                "db_table": "sage_invoice",
            },
        ),
        migrations.CreateModel(
            name="Expense",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "subtotal",
                    models.DecimalField(
                        db_comment="Sum of all item totals in the invoice",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The sum of all item totals.",
                        max_digits=10,
                        verbose_name="Subtotal",
                    ),
                ),
                (
                    "tax_percentage",
                    models.DecimalField(
                        db_comment="The percentage of tax applied to the invoice",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The tax percentage applied to the invoice.",
                        max_digits=5,
                        verbose_name="Tax Percentage",
                    ),
                ),
                (
                    "discount_percentage",
                    models.DecimalField(
                        db_comment="The percentage of discount applied to the invoice",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The discount percentage applied to the invoice.",
                        max_digits=5,
                        verbose_name="Discount Percentage",
                    ),
                ),
                (
                    "concession_percentage",
                    models.DecimalField(
                        db_comment="The percentage of concession applied to the invoice",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The concession percentage applied to the invoice.",
                        max_digits=5,
                        verbose_name="Concession Percentage",
                    ),
                ),
                (
                    "tax_amount",
                    models.DecimalField(
                        db_comment="The total tax amount calculated based on the subtotal and tax percentage",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The calculated tax amount.",
                        max_digits=10,
                        verbose_name="Tax Amount",
                    ),
                ),
                (
                    "discount_amount",
                    models.DecimalField(
                        db_comment="The total discount amount calculated based on the subtotal and discount percentage",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The calculated discount amount.",
                        max_digits=10,
                        verbose_name="Discount Amount",
                    ),
                ),
                (
                    "concession_amount",
                    models.DecimalField(
                        db_comment="The total concession amount calculated based on the subtotal and discount percentage",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The calculated concession amount.",
                        max_digits=10,
                        verbose_name="Concession Amount",
                    ),
                ),
                (
                    "total_amount",
                    models.DecimalField(
                        db_comment="The final total amount after tax and discount are applied",
                        decimal_places=2,
                        default=Decimal("0.00"),
                        help_text="The final total after applying tax and discount.",
                        max_digits=10,
                        verbose_name="Total Amount",
                    ),
                ),
                (
                    "invoice",
                    models.OneToOneField(
                        db_comment="Reference to the associated invoice",
                        help_text="The invoice associated with this total.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="total",
                        to="sage_invoice.invoice",
                        verbose_name="Invoice",
                    ),
                ),
            ],
            options={
                "verbose_name": "Expense",
                "verbose_name_plural": "Expenses",
                "db_table": "sage_expense",
            },
        ),
        migrations.CreateModel(
            name="Item",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        db_comment="Description of the invoice item",
                        help_text="Description of the item.",
                        max_length=255,
                        verbose_name="Description",
                    ),
                ),
                (
                    "quantity",
                    models.PositiveIntegerField(
                        db_comment="The quantity of the invoice item",
                        default=1,
                        help_text="The quantity of the item.",
                        verbose_name="Quantity",
                    ),
                ),
                (
                    "measurement",
                    models.CharField(
                        blank=True,
                        db_comment="measurement gavin more info about the item quantity",
                        help_text="measurement of the quantity.",
                        max_length=255,
                        null=True,
                        verbose_name="measurement",
                    ),
                ),
                (
                    "unit_price",
                    models.DecimalField(
                        db_comment="The price per unit of the invoice item",
                        decimal_places=2,
                        help_text="The price per unit of the item.",
                        max_digits=10,
                        verbose_name="Unit Price",
                    ),
                ),
                (
                    "total_price",
                    models.DecimalField(
                        db_comment="The total price calculated as quantity * unit price",
                        decimal_places=2,
                        help_text="Auto Generated total price for this item (quantity * unit price).",
                        max_digits=10,
                        verbose_name="Total Price",
                    ),
                ),
                (
                    "invoice",
                    models.ForeignKey(
                        db_comment="The associated invoice for this item",
                        help_text="The invoice associated with this item.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="items",
                        to="sage_invoice.invoice",
                        verbose_name="Invoice",
                    ),
                ),
            ],
            options={
                "verbose_name": "Item",
                "verbose_name_plural": "Items",
                "db_table": "sage_invoice_items",
            },
        ),
        migrations.CreateModel(
            name="Column",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "priority",
                    models.PositiveIntegerField(
                        db_comment="Priority of the column",
                        help_text="The priority associated with each custom column.",
                        verbose_name="Priority",
                    ),
                ),
                (
                    "column_name",
                    models.CharField(
                        db_comment="Name of the custom column",
                        help_text="The name of the custom column (e.g., 'Delivery Date', 'Warranty Period').",
                        max_length=255,
                        verbose_name="Column Name",
                    ),
                ),
                (
                    "value",
                    models.TextField(
                        db_comment="Value of the custom column",
                        help_text="The value for the custom column in the specific invoice.",
                        verbose_name="Value",
                    ),
                ),
                (
                    "invoice",
                    models.ForeignKey(
                        db_comment="invoice of the custom column",
                        help_text="The invoice associated with this custom column.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="columns",
                        to="sage_invoice.invoice",
                        verbose_name="Invoice",
                    ),
                ),
                (
                    "item",
                    models.ForeignKey(
                        db_comment="item of the custom column",
                        help_text="The item associated with this custom column.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="columns",
                        to="sage_invoice.item",
                        verbose_name="Item",
                    ),
                ),
            ],
            options={
                "verbose_name": "Column",
                "verbose_name_plural": "Columns",
                "db_table": "sage_invoice_columns",
                "ordering": ["priority"],
            },
        ),
    ]



import pytest
import secrets
from datetime import datetime
from unittest import mock
from sage_invoice.helpers.funcs import get_template_choices, generate_tracking_code
from sage_invoice.service.discovery import JinjaTemplateDiscovery


@pytest.mark.django_db
class TestHelperFunctions:

    @mock.patch("sage_invoice.helpers.funcs.JinjaTemplateDiscovery")
    @mock.patch("sage_invoice.helpers.funcs.settings")
    def test_get_template_choices_for_invoice(self, mock_settings, mock_discovery):
        """Test template choices for an invoice (non-receipt)."""
        mock_settings.SAGE_MODEL_TEMPLATE = "sage_invoice"
        mock_discovery_instance = mock_discovery.return_value
        mock_discovery_instance.SAGE_MODEL_TEMPLATE = {
            "template1": "/path/to/template1.jinja2",
            "template2": "/path/to/template2.jinja2"
        }

        choices = get_template_choices(is_receipt=False)

        assert len(choices) == 2
        assert choices[0][0] == "template1"
        assert choices[0][1] == "Template1 Template"
        assert choices[1][0] == "template2"
        assert choices[1][1] == "Template2 Template"

    @mock.patch("sage_invoice.helpers.funcs.JinjaTemplateDiscovery")
    @mock.patch("sage_invoice.helpers.funcs.settings")
    def test_get_template_choices_for_receipt(self, mock_settings, mock_discovery):
        """Test template choices for a receipt."""
        mock_settings.SAGE_MODEL_TEMPLATE = "sage_invoice"
        mock_discovery_instance = mock_discovery.return_value
        mock_discovery_instance.receipt_templates = {
            "receipt1": "/path/to/receipt1.jinja2",
            "receipt2": "/path/to/receipt2.jinja2"
        }

        choices = get_template_choices(is_receipt=True)

        assert len(choices) == 2
        assert choices[0][0] == "receipt1"
        assert choices[0][1] == "Receipt1 Template"
        assert choices[1][0] == "receipt2"
        assert choices[1][1] == "Receipt2 Template"

    def test_get_template_choices_no_templates(self):
        """Test when no templates are available."""
        with mock.patch("sage_invoice.helpers.funcs.JinjaTemplateDiscovery") as mock_discovery:
            mock_discovery_instance = mock_discovery.return_value
            mock_discovery_instance.SAGE_MODEL_TEMPLATE = {}
            mock_discovery_instance.receipt_templates = {}

            choices = get_template_choices(is_receipt=False)
            assert choices == [("", "No Templates Available")]

    @mock.patch("secrets.randbelow", return_value=1234)
    def test_generate_tracking_code(self, mock_randbelow):
        """Test the tracking code generation."""
        user_input = "INV"
        creation_date = datetime(2024, 9, 1)

        tracking_code = generate_tracking_code(user_input, creation_date)

        assert tracking_code is not None
        mock_randbelow.assert_called_once_with(8000)



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter, SearchFilter
from rest_framework.response import Response

from sage_invoice.api.helpers.versioning import HeaderVersioning
from sage_invoice.api.mixins import ErrorHandlingMixin
from sage_invoice.api.serializers import (
    ColumnSerializer,
    ExpenseSerializer,
    InvoiceSerializer,
    ItemSerializer,
)
from sage_invoice.models import Column, Expense, Invoice, Item


class InvoiceViewSet(ErrorHandlingMixin, viewsets.ModelViewSet):
    queryset = Invoice.objects.all()
    serializer_class = InvoiceSerializer
    lookup_field = "slug"
    versioning_class = HeaderVersioning
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    search_fields = (
        "title",
        "tracking_code",
        "customer_name",
    )
    filterset_fields = ("receipt", "status", "category")
    ordering_fields = ("invoice_date", "due_date")
    ordering = "-invoice_date"



from django.urls import include, path
from rest_framework.routers import DefaultRouter

from .views import (
    CategoryViewSet,
    ColumnViewSet,
    ExpenseViewSet,
    InvoiceViewSet,
    ItemViewSet,
)

router = DefaultRouter()
router.register(r"expenses", ExpenseViewSet)
router.register(r"invoices", InvoiceViewSet)
router.register(r"items", ItemViewSet)
router.register(r"categories", CategoryViewSet)
router.register(r"columns", ColumnViewSet)

urlpatterns = [
    path("", include(router.urls)),
]



# views.py
from rest_framework import viewsets

from sage_invoice.api.serializers import ColumnSerializer
from sage_invoice.models import Column


class ColumnViewSet(viewsets.ModelViewSet):
    queryset = Column.objects.all()
    serializer_class = ColumnSerializer
    lookup_field = "id"
    lookup_url_kwarg = "id"



# Generated by Django 5.1.2 on 2024-10-17 09:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sage_invoice", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="template_choice",
            field=models.CharField(
                choices=[
                    ("quotation_1", "quotation_1"),
                    ("quotation_2", "quotation_2"),
                    ("quotation_3", "quotation_3"),
                    ("quotation_4", "quotation_4"),
                    ("receipt1", "receipt1"),
                    ("receipt2", "receipt2"),
                    ("receipt3", "receipt3"),
                ],
                db_comment="Template choice for the invoice",
                help_text="The template you want for your invoice",
                max_length=20,
                verbose_name="Template choice",
            ),
        ),
    ]



import json
from decimal import Decimal

from django.core.exceptions import ValidationError
from import_export import fields, resources
from import_export.widgets import BooleanWidget, DateWidget, ForeignKeyWidget, Widget

from sage_invoice.models import Category, Column, Expense, Invoice, Item


class JSONFieldWidget(Widget):
    def clean(self, value, row=None, *args, **kwargs):
        if not value:
            return None
        try:
            return json.loads(value)
        except json.JSONDecodeError as err:
            raise ValidationError(f"Invalid JSON format in field: {value}") from err

    def render(self, value, obj=None):
        if value is None:
            return ""
        return json.dumps(value)


# ItemResource handles Invoice Items
class ItemResource(resources.ModelResource):
    invoice = fields.Field(
        column_name="invoice_id",
        attribute="invoice",
        widget=ForeignKeyWidget(Invoice, "id"),  # Use 'id' for foreign key lookup
    )

    class Meta:
        model = Item
        fields = (
            "id",
            "invoice",
            "description",
            "quantity",
            "unit_price",
            "total_price",
        )
        import_id_fields = ["id"]


# ColumnResource handles custom columns for invoices
class ColumnResource(resources.ModelResource):
    invoice = fields.Field(
        column_name="invoice_id",
        attribute="invoice",
        widget=ForeignKeyWidget(Invoice, "id"),
    )
    item = fields.Field(
        column_name="item_id",
        attribute="item",
        widget=ForeignKeyWidget(Item, "id"),
    )

    class Meta:
        model = Column
        fields = ("id", "invoice", "item", "column_name", "value", "priority")
        import_id_fields = ["id"]


class ExpenseResource(resources.ModelResource):
    invoice = fields.Field(
        column_name="invoice_id",
        attribute="invoice",
        widget=ForeignKeyWidget(Invoice, "id"),  # Use 'id' for foreign key lookup
    )

    class Meta:
        model = Expense
        fields = (
            "id",
            "invoice",
            "subtotal",
            "tax_percentage",
            "tax_amount",
            "discount_percentage",
            "discount_amount",
            "total_amount",
        )
        import_id_fields = ["id"]


# CategoryResource handles Invoice Categories
class CategoryResource(resources.ModelResource):
    class Meta:
        model = Category
        fields = (
            "id",
            "title",
            "description",
        )
        import_id_fields = ["id"]


# Main InvoiceResource to handle invoice importing/exporting
class InvoiceResource(resources.ModelResource):
    notes = fields.Field(
        column_name="notes",
        attribute="notes",
        widget=JSONFieldWidget(),
    )

    contacts = fields.Field(
        column_name="contacts",
        attribute="contacts",
        widget=JSONFieldWidget(),
    )

    invoice_date = fields.Field(
        column_name="invoice_date",
        attribute="invoice_date",
        widget=DateWidget(format="%Y-%m-%d"),
    )
    due_date = fields.Field(
        column_name="due_date",
        attribute="due_date",
        widget=DateWidget(format="%Y-%m-%d"),
    )

    receipt = fields.Field(
        column_name="receipt",
        attribute="receipt",
        widget=BooleanWidget(),
    )

    items = fields.Field()
    columns = fields.Field()
    totals = fields.Field()
    category = fields.Field()

    class Meta:
        model = Invoice
        fields = (
            "id",
            "title",
            "invoice_date",
            "customer_name",
            "customer_email",
            "status",
            "receipt",
            "contacts",
            "due_date",
            "tracking_code",
            "notes",
            "logo",
            "signature",
            "stamp",
            "template_choice",
            "category",
            "items",
            "columns",
            "totals",
        )
        import_id_fields = ["id"]

    def before_import_row(self, row, **kwargs):
        """
        Before importing, handle category, items, columns, and totals exactly like
        other fields.
        """
        # Initialize the invoice_item_map for each row import
        self.invoice_item_map = {}

        if row.get("items"):
            self.create_invoice_items(row["items"], row["id"])

        if row.get("columns"):
            self.create_invoice_columns(row["columns"], row["id"])

        if row.get("totals"):
            self.create_expenses(row["totals"], row["id"])

    def after_import_row(self, row, row_result, **kwargs):
        """After row import, assign the created or updated category to the invoice."""
        if row.get("category"):
            category_data = row["category"].split("|")
            title = category_data[0].strip()
            description = category_data[1].strip() if len(category_data) > 1 else ""

            # Create or update the Category
            category, created = Category.objects.update_or_create(
                title=title,
                defaults={"description": description},
            )

            # Now, directly set the category for the current invoice
            invoice = Invoice.objects.get(id=row["id"])
            invoice.category = category
            invoice.save()

    def create_invoice_items(self, items_data, invoice_id):
        items = items_data.split("; ")
        for item_data in items:
            try:
                description, quantity, unit_price, total_price = item_data.split("|")
                item, created = Item.objects.update_or_create(
                    invoice_id=invoice_id,
                    description=description.strip(),
                    defaults={
                        "quantity": int(quantity.strip()),
                        "unit_price": Decimal(unit_price.strip()),
                        "total_price": Decimal(total_price.strip()),
                    },
                )
                self.invoice_item_map[invoice_id] = item
            except ValueError as err:
                raise ValidationError(f"Invalid item data:{item_data}") from err

    def create_invoice_columns(self, columns_data, invoice_id):
        columns = columns_data.split("; ")
        for column_data in columns:
            try:
                column_name, value, count = column_data.split("|")
                item = self.invoice_item_map.get(invoice_id)
                if not item:
                    raise ValidationError(
                        f"No matching invoice item for invoice {invoice_id}"
                    )
                Column.objects.update_or_create(
                    invoice_id=invoice_id,
                    item=item,
                    column_name=column_name.strip(),
                    defaults={"value": value.strip(), "priority": count},
                )
            except ValueError as err:
                raise ValidationError(f"Invalid column data:{column_data}") from err

    def create_expenses(self, totals_data, invoice_id):
        try:
            (
                subtotal,
                tax_percentage,
                discount_percentage,
                concession_percentage,
            ) = totals_data.split("|")
            tax_amount = Decimal(subtotal) * (Decimal(tax_percentage) / 100)
            discount_amount = Decimal(subtotal) * (Decimal(discount_percentage) / 100)
            total_amount = (Decimal(subtotal) + tax_amount) - discount_amount

            Expense.objects.update_or_create(
                invoice_id=invoice_id,
                defaults={
                    "subtotal": Decimal(subtotal),
                    "tax_percentage": Decimal(tax_percentage),
                    "discount_percentage": Decimal(discount_percentage),
                    "tax_amount": tax_amount,
                    "discount_amount": discount_amount,
                    "total_amount": total_amount,
                },
            )
        except ValueError as err:
            raise ValidationError(f"Invalid totals data: {totals_data}") from err

    # Export related data for items, columns, totals, and contacts
    def dehydrate_items(self, invoice):
        items = Item.objects.filter(invoice=invoice)
        return "; ".join(
            [
                f"{item.description}|{item.quantity}|{item.unit_price}|{item.total_price}"
                for item in items
            ]
        )

    def dehydrate_columns(self, invoice):
        columns = Column.objects.filter(invoice=invoice)
        return "; ".join(
            [
                f"{column.column_name}|{column.value}|{column.priority}"
                for column in columns
            ]
        )

    def dehydrate_totals(self, invoice):
        totals = Expense.objects.filter(invoice=invoice)
        return "; ".join(
            [
                f"{total.subtotal}|{total.tax_percentage}|{total.discount_percentage}|{total.concession_percentage}"
                for total in totals
            ]
        )

    def dehydrate_category(self, invoice):
        """Export category data as 'title|description' format."""
        if invoice.category:
            return f"{invoice.category.title}|{invoice.category.description}"
        return ""

    def dehydrate_contacts(self, invoice):
        """Export contacts JSON field."""
        return json.dumps(invoice.contacts)



from importlib.util import find_spec

from django.conf import settings
from django.core.checks import Error, register


@register()
def check_installed_apps(app_configs, **kwargs):
    errors = []
    required_apps = [
        "sage_invoice",
    ]

    for app in required_apps:
        if app not in settings.INSTALLED_APPS:
            errors.append(
                Error(
                    f"'{app}' is missing in INSTALLED_APPS.",
                    hint=f"Add '{app}' to INSTALLED_APPS in your settings.",
                    obj=settings,
                    id=f"sage_invoice.E00{required_apps.index(app) + 1}",
                )
            )

    return errors


@register()
def check_required_libraries(app_configs, **kwargs):
    errors = []
    required_libraries = [
        "sage_invoice",
    ]

    for library in required_libraries:
        if not find_spec(library):
            errors.append(
                Error(
                    f"The required library '{library}' is not installed.",
                    hint=f"Install '{library}' via pip: pip install {library}",
                    obj=settings,
                    id=f"sage_invoice.E00{required_libraries.index(library) + 2}",
                )
            )

    return errors


@register()
def check_required_settings(app_configs, **kwargs):
    errors = []

    required_settings = [
        "SAGE_MODEL_PREFIX",
        "SAGE_MODEL_TEMPLATE",
    ]

    for setting in required_settings:
        if not hasattr(settings, setting):
            errors.append(
                Error(
                    f"The required setting '{setting}' is not defined.",
                    hint=f"Define '{setting}' in your settings.py.",
                    obj=settings,
                    id=f"sage_invoice.E00{required_settings.index(setting) + 3}",
                )
            )
    return errors



from decimal import Decimal

from django.db import models
from django.utils.translation import gettext_lazy as _


class Expense(models.Model):
    subtotal = models.DecimalField(
        verbose_name=_("Subtotal"),
        max_digits=10,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The sum of all item totals."),
        db_comment="Sum of all item totals in the invoice",
    )
    tax_percentage = models.DecimalField(
        verbose_name=_("Tax Percentage"),
        max_digits=5,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The tax percentage applied to the invoice."),
        db_comment="The percentage of tax applied to the invoice",
    )
    discount_percentage = models.DecimalField(
        verbose_name=_("Discount Percentage"),
        max_digits=5,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The discount percentage applied to the invoice."),
        db_comment="The percentage of discount applied to the invoice",
    )
    concession_percentage = models.DecimalField(
        verbose_name=_("Concession Percentage"),
        max_digits=5,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The concession percentage applied to the invoice."),
        db_comment="The percentage of concession applied to the invoice",
    )
    tax_amount = models.DecimalField(
        verbose_name=_("Tax Amount"),
        max_digits=10,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The calculated tax amount."),
        db_comment="The total tax amount calculated based on the subtotal and tax percentage",
    )
    discount_amount = models.DecimalField(
        verbose_name=_("Discount Amount"),
        max_digits=10,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The calculated discount amount."),
        db_comment="The total discount amount calculated based on the subtotal and discount percentage",
    )
    concession_amount = models.DecimalField(
        verbose_name=_("Concession Amount"),
        max_digits=10,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The calculated concession amount."),
        db_comment="The total concession amount calculated based on the subtotal and discount percentage",
    )
    total_amount = models.DecimalField(
        verbose_name=_("Total Amount"),
        max_digits=10,
        decimal_places=2,
        default=Decimal("0.00"),
        help_text=_("The final total after applying tax and discount."),
        db_comment="The final total amount after tax and discount are applied",
    )
    invoice = models.OneToOneField(
        "Invoice",
        verbose_name=_("Invoice"),
        on_delete=models.CASCADE,
        related_name="expense",
        help_text=_("The invoice associated with this total."),
        db_comment="Reference to the associated invoice",
    )

    def __str__(self):
        return f"Expense {self.invoice.pk}"

    def __repr__(self):
        return f"<Expense(subtotal={self.subtotal}, total_amount={self.total_amount})>"

    class Meta:
        verbose_name = _("Expense")
        verbose_name_plural = _("Expenses")
        db_table = "sage_expense"



import logging

from django.core.exceptions import ValidationError
from django.db import IntegrityError, OperationalError, transaction
from django.db.models.signals import post_save
from django.dispatch import receiver

from sage_invoice.service.total import ExpenseService

from .models import Expense, Invoice

logger = logging.getLogger()


@receiver(post_save, sender=Invoice)
def update_invoice_total_on_save(sender, instance, created, **kwargs):
    def recalculate_total():
        invoice_total, _ = Expense.objects.get_or_create(invoice=instance)
        ExpenseService().calculate_and_save(invoice_total)

    try:
        transaction.on_commit(recalculate_total)
    except IntegrityError as error:
        logger.error(
            "Integrity error occurred for invoice %s: %s", instance.title, error
        )
    except OperationalError as error:
        logger.error(
            "Operational error in database for invoice %s: %s", instance.title, error
        )
    except ValidationError as error:
        logger.error("Validation error for invoice %s: %s", instance.title, error)
    except Exception as error:
        logger.exception(
            "Unexpected error occurred for invoice %s: %s", instance.title, error
        )



from django.db import models
from django.utils.translation import gettext_lazy as _


class Item(models.Model):
    description = models.CharField(
        max_length=255,
        verbose_name=_("Description"),
        help_text=_("Description of the item."),
        db_comment="Description of the invoice item",
    )
    quantity = models.PositiveIntegerField(
        verbose_name=_("Quantity"),
        null=True,
        blank=True,
        help_text=_("The quantity of the item."),
        db_comment="The quantity of the invoice item",
    )
    measurement = models.CharField(
        max_length=255,
        verbose_name=_("measurement"),
        help_text=_("measurement of the quantity."),
        db_comment="measurement gavin more info about the item quantity",
        null=True,
        blank=True,
    )
    unit_price = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        verbose_name=_("Unit Price"),
        help_text=_("The price per unit of the item."),
        db_comment="The price per unit of the invoice item",
    )
    total_price = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        verbose_name=_("Total Price"),
        help_text=_(
            "Auto Generated total price for this item (quantity * unit price)."
        ),
        db_comment="The total price calculated as quantity * unit price",
    )
    invoice = models.ForeignKey(
        "Invoice",
        on_delete=models.CASCADE,
        related_name="items",
        verbose_name=_("Invoice"),
        help_text=_("The invoice associated with this item."),
        db_comment="The associated invoice for this item",
    )

    def save(self, *args, **kwargs):
        if self.quantity:
            self.total_price = self.quantity * self.unit_price
        else:
            self.total_price = self.unit_price
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.description} - {self.quantity} x {self.unit_price}"

    def __repr__(self):
        return f"Items> {self.description} - {self.quantity} x {self.unit_price}"

    class Meta:
        verbose_name = _("Item")
        verbose_name_plural = _("Items")
        db_table = "sage_invoice_items"



import logging
import uuid

from django.utils.timezone import now
from rest_framework import status
from rest_framework.exceptions import (
    APIException,
    NotAcceptable,
    NotFound,
    ValidationError,
)
from rest_framework.response import Response
from rest_framework.views import exception_handler

# Get the logger instance
logger = logging.getLogger(__name__)


class ErrorHandlingMixin:
    """
    Mixin to provide consistent error handling across APIs,
    including structured error responses, logging, and trace IDs.
    """

    def handle_exception(self, exc):
        """
        Override the default DRF exception handling to provide
        a structured error response with logging and trace IDs.
        """
        # Call the default DRF exception handler first to get the default response
        response = exception_handler(exc, self.get_exception_handler_context())

        # Generate a unique trace ID for this request
        trace_id = str(uuid.uuid4())

        # Path from the current request
        path = self.request.path if hasattr(self, "request") else "unknown"

        # Default error structure
        error_response = {
            "code": "SERVER_ERROR",
            "message": "An unexpected error occurred.",
            "details": [],
            "path": path,
            "traceId": trace_id,
            "timestamp": now().isoformat(),
        }

        # Handle specific exceptions for better detail
        if isinstance(exc, ValidationError):
            error_response["code"] = "VALIDATION_ERROR"
            error_response["message"] = "Invalid input parameters"
            error_response["details"] = [
                {"field": field, "message": msg[0] if isinstance(msg, list) else msg}
                for field, msg in exc.detail.items()
            ]
            status_code = status.HTTP_400_BAD_REQUEST

        elif isinstance(exc, NotFound):
            error_response["code"] = "NOT_FOUND"
            error_response["message"] = str(exc)
            status_code = status.HTTP_404_NOT_FOUND

        elif isinstance(exc, NotAcceptable):
            error_response["code"] = "NOT_ACCEPTABLE"
            error_response["message"] = str(exc)
            status_code = status.HTTP_406_NOT_ACCEPTABLE

        elif isinstance(exc, APIException):
            error_response["code"] = exc.get_codes() if exc.get_codes() else "API_ERROR"
            error_response["message"] = str(exc)
            status_code = exc.status_code

        else:
            # If response is None, this means DRF didn't handle the exception, so we do it ourselves.
            status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        # If DRF handled the exception, override its response data with the structured error
        if response is not None:
            response.data = error_response
            response.status_code = status_code
        else:
            # Create a custom response if DRF couldn't handle the exception
            response = Response(error_response, status=status_code)

        # Log the error details with trace ID for correlation
        logger.error(
            f"Error occurred [traceId={trace_id}] at {path}: {exc}", exc_info=True
        )

        return response

    def get_exception_handler_context(self):
        """
        This method returns the context that will be passed to the DRF exception
        handler.

        It's useful to provide additional request-related context (like view, request, args).
        """
        return {
            "view": self,
            "request": self.request,
        }



# serializers.py
from rest_framework import serializers

from sage_invoice.models import Category


class CategorySerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="category-detail", lookup_field="slug"
    )
    invoices = serializers.HyperlinkedRelatedField(
        many=True, view_name="invoice-detail", lookup_field="slug", read_only=True
    )

    class Meta:
        model = Category
        fields = ("url", "title", "description", "invoices")
        extra_kwargs = {
            "url": {"lookup_field": "slug"},
        }



import os

from django.apps import apps
from django.conf import settings


class JinjaTemplateDiscovery:
    def __init__(self):
        self.default_template_dir = "default_invoices"  # inside package
        self.sage_template_dir = (
            settings.SAGE_MODEL_TEMPLATE
        )  # custom templates folder name
        self.sage_template_prefix = settings.SAGE_MODEL_PREFIX  # custom template prefix

    def get_default_templates(self):
        """Return a list of default templates found in the package's template directory."""
        # Use Django's template loading mechanisms to discover default templates.
        default_path = os.path.join(
            apps.get_app_config("sage_invoice").path,
            "templates",
            self.default_template_dir,
        )
        return self._find_templates_in_directory(default_path)

    def get_custom_templates(self):
        """Return a list of custom templates found in each app's `templates/` directory."""
        template_choices = []
        for app_config in apps.get_app_configs():
            template_dir = os.path.join(
                app_config.path, "templates", self.sage_template_dir
            )
            if os.path.exists(template_dir):
                template_choices.extend(
                    self._find_templates_in_directory(
                        template_dir, self.sage_template_prefix
                    )
                )
        return template_choices

    def _find_templates_in_directory(self, directory, prefix=None):
        """
        Helper method to find .jinja2 files in a directory, optionally filtering by
        prefix, and return the filenames without the .jinja2 extension.
        """
        if not os.path.exists(directory):
            return []

        templates = []
        for filename in os.listdir(directory):
            if filename.endswith(".jinja2") and (
                not prefix or filename.startswith(prefix)
            ):
                templates.append(filename)

        # Remove the .jinja2 extension from the filenames
        filenames = list(map(lambda x: x.replace(".jinja2", ""), templates))

        return filenames



from django import template

register = template.Library()


@register.filter
def get_item(dictionary, key):
    return dictionary.get(key, "")


@register.filter
def subtract(value, arg):
    try:
        return float(value) - float(arg)
    except (ValueError, TypeError):
        return ""


@register.filter
def split_by_period(value):
    """Splits a string by '.' and returns a list of sentences."""
    if isinstance(value, str):
        return [sentence.strip() for sentence in value.split(".") if sentence.strip()]
    return []



from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django_jsonform.models.fields import JSONField
from sage_tools.mixins.models import TitleSlugMixin

from sage_invoice.helpers.choice import Currency, InvoiceStatus
from sage_invoice.helpers.funcs import generate_tracking_code, get_template_choices


class Invoice(TitleSlugMixin):
    invoice_date = models.DateField(
        verbose_name=_("Invoice Date"),
        help_text=_("The date when the invoice was created."),
        db_comment="Invoice date created",
    )
    customer_name = models.CharField(
        max_length=255,
        verbose_name=_("Customer Name"),
        help_text=_("The name of the customer."),
        db_comment="Customer name created",
    )
    tracking_code = models.CharField(
        max_length=255,
        verbose_name=_("Tracking Code"),
        help_text=_(
            "Enter the first 3-4 characters of the tracking code. The full code will be auto-generated as <prefix> + <date> + <random number>."
        ),
        db_comment="Tracking code created",
    )
    contacts = JSONField(
        verbose_name="Customer Contacts",
        blank=True,
        null=True,
        schema={
            "type": "object",
            "properties": {
                "Contact Info": {
                    "oneOf": [
                        {
                            "type": "object",
                            "title": "Phone",
                            "properties": {
                                "phone": {
                                    "type": "string",
                                    "pattern": "^[0-9]+$",
                                    "title": "Phone",
                                    "placeholder": "1234567890",
                                }
                            },
                        },
                        {
                            "type": "object",
                            "title": "Email",
                            "properties": {
                                "email": {
                                    "type": "string",
                                    "format": "email",
                                    "title": "Email",
                                    "placeholder": "you@example.com",
                                }
                            },
                        },
                    ]
                }
            },
        },
    )
    status = models.CharField(
        max_length=50,
        choices=InvoiceStatus,
        verbose_name=_("Status"),
        help_text=_("The current status of the invoice (Paid, Unpaid)."),
        db_comment="Current status of the invoice (Paid, Unpaid)",
    )
    receipt = models.BooleanField(
        verbose_name=_("Receipt"),
        default=False,
        help_text=_("Is this a receipt or an invoice"),
        db_comment="Check if this invoice is a receipt",
    )
    notes = JSONField(
        verbose_name=_("Notes"),
        blank=True,
        null=True,
        help_text=(
            """
            You can add any number of custom fields dynamically,
            such as'Terms & Conditions', 'Technology Tips', etc.
            """
        ),
        db_comment=("This field stores additional dynamic content in JSON format. "),
        schema={
            "type": "array",
            "title": "Additional Fields",
            "items": {
                "type": "object",
                "title": "Field",
                "properties": {
                    "label": {"type": "string", "title": "Field Name"},
                    "content": {
                        "type": "string",
                        "title": "Field Content",
                        "widget": "textarea",
                    },
                },
            },
        },
    )
    category = models.ForeignKey(
        "Category",
        on_delete=models.CASCADE,
        related_name="invoices",
        null=True,
        blank=True,
        verbose_name=_("Category"),
        help_text=_("The category associated with this invoice."),
        db_comment="Category associated with this invoice",
    )
    currency = models.CharField(
        max_length=10,
        verbose_name="Currency",
        choices=Currency.choices,
        default=Currency.USD,
        help_text=_("Currency of unit price"),
        db_comment="Which currency is this item ",
    )
    due_date = models.DateField(
        verbose_name=_("Due Date"),
        help_text=_("The date by which the invoice should be paid."),
        db_comment="Due date of the invoice",
    )
    logo = models.ImageField(
        upload_to="invoices/logos/",
        blank=True,
        null=True,
        verbose_name=_("Logo"),
        help_text=_("The logo displayed on the invoice."),
        db_comment="Logo displayed on the invoice",
    )
    signature = models.ImageField(
        upload_to="invoices/signatures/",
        blank=True,
        null=True,
        verbose_name=_("Signature"),
        help_text=_("The signature image for the invoice."),
        db_comment="Signature image for the invoice",
    )
    stamp = models.ImageField(
        upload_to="invoices/stamps/",
        blank=True,
        null=True,
        verbose_name=_("Stamp"),
        help_text=_("The stamp image for the invoice."),
        db_comment="Stamp image for the invoice",
    )
    template_choice = models.CharField(
        verbose_name=_("Template choice"),
        max_length=20,
        choices=get_template_choices(),
        help_text=_("The template you want for your invoice"),
        db_comment="Template choice for the invoice",
    )

    def clean(self):
        if self.due_date < self.invoice_date:
            raise ValidationError(_("Due Date must be later than Invoice Date."))
        if len(self.tracking_code) <= 10:
            self.tracking_code = generate_tracking_code(
                self.tracking_code, self.invoice_date
            )

    def get_absolute_url(self):
        return reverse("invoice_detail", kwargs={"slug": self.slug})

    class Meta:
        verbose_name = _("Invoice")
        verbose_name_plural = _("Invoices")
        db_table = "sage_invoice"

    def __str__(self):
        return f"{self.title}"

    def __repr__(self):
        return f"Invoice> {self.title}>"



from django.db import models
from django.utils.translation import gettext_lazy as _
from sage_tools.mixins.models import TitleSlugDescriptionMixin


class Category(TitleSlugDescriptionMixin):
    def __str__(self):
        return f"{self.title}"

    def __repr__(self):
        return f"{self.title}"

    class Meta:
        verbose_name = _("Category")
        verbose_name_plural = _("Categories ")
        db_table = "sage_invoice_cat"



import pytest
from unittest.mock import patch, MagicMock
from decimal import Decimal
from sage_invoice.models import Invoice, Item, Column, Expense
from sage_invoice.resource import InvoiceResource,Category,ItemResource, ExpenseResource


@pytest.mark.django_db
class TestInvoiceResource:

    @pytest.fixture
    def invoice(self, db):
        return Invoice.objects.create(
            title="Test Invoice",
            invoice_date="2024-09-10",
            customer_name="John Doe",
            status="unpaid",
            tracking_code="INV-20240910",
            due_date="2024-09-20"  # Add the due_date here
        )

    @pytest.fixture
    def invoice_item(self, invoice):
        return Item.objects.create(
            invoice=invoice,
            description="Item 1",
            quantity=2,
            unit_price=Decimal("100.00"),
            total_price=Decimal("200.00"),
        )

    @pytest.fixture
    def invoice_column(self, invoice, invoice_item):
        return Column.objects.create(
            invoice=invoice,
            item=invoice_item,
            column_name="Column 1",
            value="Value 1",
            priority=1,
        )

    @pytest.fixture
    def expense(self, invoice):
        return Expense.objects.create(
            invoice=invoice,
            subtotal=Decimal("1000.00"),
            tax_percentage=Decimal("10.00"),
            tax_amount=Decimal("100.00"),
            discount_percentage=Decimal("5.00"),
            discount_amount=Decimal("50.00"),
            total_amount=Decimal("1050.00"),
        )


    @patch.object(Expense, 'objects')
    def test_create_expenses(self, mock_objects, invoice):
        """Test the create_expenses method."""
        mock_objects.update_or_create = MagicMock()
        resource = InvoiceResource()

        resource.create_expenses("1000.00|10.00|5.00|0", invoice.id)

        mock_objects.update_or_create.assert_called_once_with(
            invoice_id=invoice.id,
            defaults={
                "subtotal": Decimal("1000.00"),
                "tax_percentage": Decimal("10.00"),
                "discount_percentage": Decimal("5.00"),
                "tax_amount": Decimal("100.00"),
                "discount_amount": Decimal("50.00"),
                "total_amount": Decimal("1050.00"),
            }
        )

    def test_dehydrate_items(self, invoice, invoice_item):
        """Test the dehydrate_items method."""
        resource = InvoiceResource()
        result = resource.dehydrate_items(invoice)
        assert result is not None

    def test_dehydrate_columns(self, invoice, invoice_column):
        """Test the dehydrate_columns method."""
        resource = InvoiceResource()
        result = resource.dehydrate_columns(invoice)
        assert result == "Column 1|Value 1|1"

    def test_dehydrate_totals(self, invoice, expense):
        """Test the dehydrate_totals method."""
        resource = InvoiceResource()
        result = resource.dehydrate_totals(invoice)
        assert result == "1000.00|10.00|5.00|0.00"

    def test_dehydrate_contacts(self, invoice):
        """Test the dehydrate_contacts method."""
        invoice.contacts = [{"email": "test@example.com", "phone": "123456789"}]
        invoice.save()

        resource = InvoiceResource()
        result = resource.dehydrate_contacts(invoice)
        assert result == '[{"email": "test@example.com", "phone": "123456789"}]'


@pytest.mark.django_db
class TestItemResource:

    def test_meta(self):
        """Test that ItemResource meta is set correctly."""
        resource = ItemResource()
        assert resource.Meta.model == Item
        assert resource.Meta.fields == ("id", "invoice", "description", "quantity", "unit_price", "total_price")
        assert resource.Meta.import_id_fields == ["id"]


@pytest.mark.django_db
class TestExpenseResource:

    def test_meta(self):
        """Test that ExpenseResource meta is set correctly."""
        resource = ExpenseResource()
        assert resource.Meta.model == Expense
        assert resource.Meta.fields == ("id", "invoice", "subtotal", "tax_percentage", "tax_amount", "discount_percentage", "discount_amount", "total_amount")
        assert resource.Meta.import_id_fields == ["id"]

import pytest
from unittest.mock import patch, MagicMock
from sage_invoice.models import Invoice, Category
from sage_invoice.resource import InvoiceResource

@pytest.mark.django_db
class TestInvoiceResourceWithCategory:

    @pytest.fixture
    def category(self):
        """Fixture to create a category."""
        return Category.objects.create(title="Consulting", description="Consulting Services")

    @pytest.fixture
    def invoice_with_category(self, category):
        """Fixture to create an invoice with a category."""
        return Invoice.objects.create(
            title="Invoice with Category",
            invoice_date="2024-09-11",
            customer_name="Jane Doe",
            status="paid",
            due_date="2024-09-30",
            category=category,
        )

    @pytest.fixture
    def invoice_without_category(self):
        """Fixture to create an invoice without a category."""
        return Invoice.objects.create(
            title="Invoice without Category",
            invoice_date="2024-09-11",
            customer_name="John Doe",
            status="unpaid",
            due_date="2024-09-30"
        )

    def test_dehydrate_category(self, invoice_with_category):
        """Test the dehydrate_category method when a category exists."""
        resource = InvoiceResource()
        result = resource.dehydrate_category(invoice_with_category)
        assert result == "Consulting|Consulting Services"

    def test_dehydrate_category_no_category(self, invoice_without_category):
        """Test the dehydrate_category method when no category is assigned."""
        resource = InvoiceResource()
        result = resource.dehydrate_category(invoice_without_category)
        assert result == ""

    def test_import_category_empty(self, invoice_without_category):
        """Test that an empty category field does not assign a category."""
        resource = InvoiceResource()

        row = {
            "id": invoice_without_category.id,
            "category": ""  # Empty category field
        }

        # Simulate the after_import_row method
        resource.after_import_row(row, None)

        # Ensure that the category is not assigned
        invoice = Invoice.objects.get(id=row["id"])
        assert invoice.category is None



from rest_framework import viewsets

from sage_invoice.api.helpers.versioning import HeaderVersioning
from sage_invoice.api.mixins import ErrorHandlingMixin
from sage_invoice.api.serializers import CategorySerializer
from sage_invoice.models import Category


class CategoryViewSet(ErrorHandlingMixin, viewsets.ModelViewSet):
    queryset = Category.objects.all()
    serializer_class = CategorySerializer

    versioning_class = HeaderVersioning

    def get_serializer_class(self):
        version = getattr(self.request, "version", None)

        if version == "1.0":
            return CategorySerializer

        return super().get_serializer_class()



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from sage_invoice.models import Category


@admin.register(Category)
class CategoryAdmin(admin.ModelAdmin):
    admin_priority = 2
    list_display = ("title", "description")
    search_fields = ("title", "description")
    readonly_fields = ("slug",)
    ordering = ("title",)
    save_on_top = True

    fieldsets = (
        (
            _("Category Details"),
            {
                "fields": (
                    "title",
                    "slug",
                    "description",
                ),
                "description": _(
                    "Details of the invoice category including title and description."
                ),
            },
        ),
    )



import os
import secrets
from datetime import datetime

from django.conf import settings

from sage_invoice.service.discovery import JinjaTemplateDiscovery


def get_template_choices():
    """
    Returns a combined list of default and custom templates, formatted for use in a
    model's choices field.
    """
    discovery = JinjaTemplateDiscovery()
    default_templates = discovery.get_default_templates()
    custom_templates = discovery.get_custom_templates()

    choices = [(template, template) for template in default_templates]
    choices += [(template, template) for template in custom_templates]

    if not choices:
        return [("", "No templates available")]

    return choices


def generate_tracking_code(user_input: str, creation_date: datetime) -> str:
    """Generate a unique tracking code based on user input and the creation
    date.

    Returns:
        str: A unique tracking code.
    """
    date_str = creation_date.strftime("%Y%m%d")
    random_number = secrets.randbelow(8000) + 1000
    tracking_code = f"{user_input}-{date_str}-{random_number}"
    return tracking_code



import pytest
from unittest import mock
from django.db import IntegrityError, OperationalError, transaction
from sage_invoice.models import Invoice, Expense
from sage_invoice.service.total import ExpenseService
from sage_invoice.signals import update_invoice_total_on_save
from django.core.exceptions import ValidationError


@pytest.mark.django_db
class TestInvoiceSignals:

    @pytest.fixture
    def invoice(self, db):
        """Fixture to create a real Invoice instance."""
        return Invoice.objects.create(
            title="Test Invoice",
            invoice_date="2024-09-01",
            customer_name="John Doe",
            status="unpaid",
            due_date="2024-09-15",
        )

    @mock.patch.object(ExpenseService, "calculate_and_save")
    @mock.patch("sage_invoice.signals.transaction.on_commit")
    def test_update_invoice_total_on_save_success(self, mock_on_commit, mock_calculate_and_save, invoice):
        """Test that the signal correctly calculates and saves the invoice total."""
        Expense.objects.create(invoice=invoice, subtotal=100, total_amount=100)

        # Trigger the signal
        update_invoice_total_on_save(Invoice, invoice, created=False)

        # Verify that the recalculate_total function is registered with on_commit
        mock_on_commit.assert_called_once()

        # Call the recalculate_total function manually
        recalculate_total_fn = mock_on_commit.call_args[0][0]
        recalculate_total_fn()  # Simulate transaction commit

        mock_calculate_and_save.assert_called_once()

    @mock.patch.object(ExpenseService, "calculate_and_save")
    @mock.patch("sage_invoice.signals.transaction.on_commit")
    def test_update_invoice_total_on_save_transaction_error(self, mock_on_commit, mock_calculate_and_save, invoice):
        """Test that transaction errors are handled correctly."""
        mock_on_commit.side_effect = OperationalError("Transaction failed")

        # Trigger the signal
        update_invoice_total_on_save(Invoice, invoice, created=False)

        mock_calculate_and_save.assert_not_called()

    @mock.patch.object(ExpenseService, "calculate_and_save")
    @mock.patch("sage_invoice.signals.transaction.on_commit")
    def test_update_invoice_total_on_save_integrity_error(self, mock_on_commit, mock_calculate_and_save, invoice):
        """Test that integrity errors are handled correctly."""
        mock_on_commit.side_effect = IntegrityError("Integrity failed")

        # Trigger the signal
        update_invoice_total_on_save(Invoice, invoice, created=False)

        mock_calculate_and_save.assert_not_called()

    @mock.patch.object(ExpenseService, "calculate_and_save")
    @mock.patch("sage_invoice.signals.transaction.on_commit")
    def test_update_invoice_total_on_save_validation_error(self, mock_on_commit, mock_calculate_and_save, invoice):
        """Test that validation errors are handled correctly."""
        mock_on_commit.side_effect = ValidationError("Validation failed")

        # Trigger the signal
        update_invoice_total_on_save(Invoice, invoice, created=False)

        mock_calculate_and_save.assert_not_called()



from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.core.exceptions import PermissionDenied
from django.http import JsonResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.views import View
from django.views.generic import DetailView, TemplateView

from sage_invoice.helpers.funcs import get_template_choices
from sage_invoice.models import Invoice
from sage_invoice.service.invoice_create import QuotationService


class InvoiceDetailView(LoginRequiredMixin, UserPassesTestMixin, DetailView):
    model = Invoice
    context_object_name = "invoice"
    queryset = Invoice.objects.select_related("category", "expense").prefetch_related(
        "items", "columns"
    )

    def test_func(self):
        # Ensure that the user is a staff member
        return self.request.user.is_staff

    def handle_no_permission(self):
        # If the user is not authenticated, it will redirect to login
        # If the user is authenticated but lacks permission, it raises PermissionDenied
        if self.request.user.is_authenticated:
            raise PermissionDenied(self.permission_denied_message)
        return super().handle_no_permission()

    def get_template_names(self):
        # Dynamically select the template based on the invoice type
        invoice = self.get_object()
        return [f"{invoice.template_choice}.html"]

    def get_context_data(self, **kwargs):
        # Add additional context using the service class
        context = super().get_context_data(**kwargs)
        invoice = self.get_object()
        context["customer_email"] = invoice.contacts.get("Contact Info").get(
            "email", None
        )
        context["customer_phone"] = invoice.contacts.get("Contact Info").get(
            "phone", None
        )
        return context


class TemplateChoiceView(View):
    def get(self, request, *args, **kwargs):
        is_receipt = self.kwargs.get("check")
        if "T" in is_receipt:
            is_receipt = True
        else:
            is_receipt = False
        choices = get_template_choices(is_receipt=is_receipt)
        formatted_choices = [
            {"value": choice[0], "label": choice[1]} for choice in choices
        ]
        return JsonResponse(formatted_choices, safe=False)


class GenerateInvoicesView(TemplateView):
    """
    This view generates multiple invoices and returns their rendered HTML for PDF
    generation.
    """

    def get(self, request, *args, **kwargs):
        invoice_ids = request.GET.get("invoice_ids", "")
        invoice_ids = invoice_ids.split(",") if invoice_ids else []

        if not invoice_ids:
            return JsonResponse({"error": "No invoice IDs provided"}, status=400)

        invoices_data = []

        for invoice_id in invoice_ids:
            invoice = Invoice.objects.filter(id=invoice_id).first()
            if not invoice:
                continue

            service = QuotationService()
            rendered_content = service.render_context(invoice)

            # Determine the template based on invoice type
            if invoice.receipt:
                template_name = f"receipt{invoice.template_choice}.html"
            else:
                template_name = f"quotation{invoice.template_choice}.html"

            rendered_html = render_to_string(template_name, rendered_content)

            invoices_data.append(
                {
                    "id": invoice.id,
                    "title": invoice.title or f"Invoice_{invoice.id}",
                    "rendered_html": rendered_html,
                }
            )

        return JsonResponse({"invoices": invoices_data})


class DownloadInvoicesView(View):
    def get(self, request, *args, **kwargs):
        # Get the invoice IDs from the query parameters
        invoice_ids = request.GET.get("invoice_ids", "")

        context = {"invoice_ids": invoice_ids}

        return render(request, "download_invoices.html", context)



from django.contrib import admin
from django.shortcuts import redirect
from django.urls import reverse


@admin.action(description="Download selected invoice as PDF")
def export_pdf(modeladmin, request, queryset):
    invoice_ids = queryset.values_list("id", flat=True)
    invoice_ids_str = ",".join(map(str, invoice_ids))
    url = f"{reverse('download_invoices')}?invoice_ids={invoice_ids_str}"
    return redirect(url)



from .category import Category
from .column import Column
from .expense import Expense
from .invoice import Invoice
from .item import Item

__all__ = ["Invoice", "Column", "Item", "Expense", "Category"]



from django.db import models
from django.utils.translation import gettext_lazy as _


class Column(models.Model):
    priority = models.PositiveIntegerField(
        verbose_name=_("Priority"),
        help_text=_("The priority associated with each custom column."),
        db_comment="Priority of the column",
    )
    column_name = models.CharField(
        max_length=255,
        verbose_name=_("Column Name"),
        help_text=_(
            "The name of the custom column (e.g., 'Delivery Date', 'Warranty Period')."
        ),
        db_comment="Name of the custom column",
    )
    value = models.TextField(
        verbose_name=_("Value"),
        help_text=_("The value for the custom column in the specific invoice."),
        db_comment="Value of the custom column",
    )
    invoice = models.ForeignKey(
        "Invoice",
        on_delete=models.CASCADE,
        related_name="columns",
        verbose_name=_("Invoice"),
        help_text=_("The invoice associated with this custom column."),
        db_comment="invoice of the custom column",
    )
    item = models.ForeignKey(
        "Item",
        on_delete=models.CASCADE,
        related_name="columns",
        verbose_name=_("Item"),
        help_text=_("The item associated with this custom column."),
        db_comment="item of the custom column",
    )

    def __str__(self):
        return f"{self.column_name}"

    def __repr__(self) -> str:
        return f"Column> {self.column_name}"

    class Meta:
        verbose_name = _("Column")
        verbose_name_plural = _("Columns")
        db_table = "sage_invoice_columns"
        ordering = ["priority"]



from decimal import Decimal


class ExpenseService:
    """Service class to handle calculations and saving for Expense."""

    def calculate_and_save(self, invoice_total, *args, **kwargs):
        """Calculate the subtotal, tax amount, discount amount, and total
        amount, and then save the instance.
        """
        # Calculate subtotal from related items, ensure it's not None
        invoice_total.subtotal = sum(
            item.total_price for item in invoice_total.invoice.items.all()
        )

        # Convert tax and discount percentages to Decimal
        tax_percentage = Decimal(invoice_total.tax_percentage)
        discount_percentage = Decimal(invoice_total.discount_percentage)
        concession_percentage = Decimal(invoice_total.concession_percentage)
        # Calculate tax amount
        invoice_total.tax_amount = invoice_total.subtotal * (
            tax_percentage / Decimal(100)
        )

        # Calculate discount amount
        invoice_total.discount_amount = invoice_total.subtotal * (
            discount_percentage / Decimal(100)
        )
        invoice_total.concession_amount = invoice_total.subtotal * (
            concession_percentage / Decimal(100)
        )
        # Calculate total amount
        invoice_total.total_amount = (
            invoice_total.subtotal
            + invoice_total.tax_amount
            - invoice_total.discount_amount
            - invoice_total.concession_amount
        )

        # Save the Expense instance using the standard save method
        invoice_total.save(*args, **kwargs)



from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from import_export.admin import ImportExportModelAdmin

from sage_invoice.admin.actions import export_pdf
from sage_invoice.models import Column, Expense, Invoice, Item
from sage_invoice.resource import InvoiceResource


class ItemInline(admin.TabularInline):
    model = Item
    extra = 0
    min_num = 1
    readonly_fields = ("total_price",)


class ColumnInline(admin.StackedInline):
    model = Column
    extra = 1


class ExpenseInline(admin.TabularInline):
    model = Expense
    extra = 1
    readonly_fields = (
        "subtotal",
        "tax_amount",
        "discount_amount",
        "total_amount",
        "concession_amount",
    )


@admin.register(Invoice)
class InvoiceAdmin(ImportExportModelAdmin, admin.ModelAdmin):
    resource_class = InvoiceResource
    admin_priority = 1
    list_display = ("title", "invoice_date", "customer_name", "status")
    search_fields = ("customer_name", "status", "customer_email")
    save_on_top = True
    list_filter = ("status", "invoice_date", "category")
    ordering = ("-invoice_date",)
    autocomplete_fields = ("category",)
    readonly_fields = ("slug",)
    actions = [export_pdf]

    class Media:
        js = ("assets/js/invoice_admin.js",)

    def get_fieldsets(self, request, obj=None):
        fieldsets = (
            (
                _("Invoice Details"),
                {
                    "fields": (
                        "title",
                        "slug",
                        "invoice_date",
                        "tracking_code",
                        "due_date",
                        "customer_name",
                        "contacts",
                        "category",
                        "receipt",
                    ),
                    "description": _(
                        """
                        Basic details of the invoice including title, date,
                        and customer information."""
                    ),
                },
            ),
            (
                _("Status & Currency"),
                {
                    "fields": ("status", "notes", "currency"),
                    "description": _(
                        "Current status of the invoice and any additional notes."
                    ),
                },
            ),
        )
        fieldsets += (
            (
                _("Design Elements"),
                {
                    "fields": (
                        "logo",
                        "signature",
                        "stamp",
                        "template_choice",
                    ),
                    "description": _(
                        "Design-related elements like logo, and template choice."
                    ),
                },
            ),
        )

        return fieldsets

    inlines = [ItemInline, ColumnInline, ExpenseInline]

    def get_inline_instances(self, request, obj=None):
        inlines = []
        if obj and obj.pk:
            inlines = [
                ItemInline(self.model, self.admin_site),
                ColumnInline(self.model, self.admin_site),
                ExpenseInline(self.model, self.admin_site),
            ]
        else:
            inlines = [
                ItemInline(self.model, self.admin_site),
                ExpenseInline(self.model, self.admin_site),
            ]
        return inlines



import logging
import os
from typing import Any, Dict

from django.conf import settings
from django.db.models import QuerySet
from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2.exceptions import TemplateNotFound

from sage_invoice.models import Column

from .discovery import JinjaTemplateDiscovery

logger = logging.getLogger(__name__)


class QuotationService:
    """Service class to handle the generation and rendering of quotations."""

    def __init__(self) -> None:
        """Initialize the QuotationService with the template discovery and
        Jinja2.
        """
        logger.info("Initializing QuotationService")
        self.template_discovery = JinjaTemplateDiscovery()
        self.env = Environment(
            loader=FileSystemLoader(self.template_discovery.sage_template_dir),
            autoescape=select_autoescape(
                ["html", "xml"]
            ),  # Enable autoescape for HTML and XML templates
        )
        logger.info(
            "Template discovery set to directory: %s",
            self.template_discovery.sage_template_dir,
        )

    def render_quotation(self, queryset: QuerySet) -> str:
        """Render the quotation for the given queryset.

        Args:
            queryset (QuerySet): A queryset containing the invoice to render.

        Returns:
            str: The rendered HTML of the quotation.

        Raises:
            TemplateNotFound: If the selected template is not found.
        """
        logger.info("Rendering quotation")
        invoice = queryset.first()
        context = self.render_context(queryset)
        is_receipt = invoice.receipt
        template_number = "".join(filter(str.isdigit, invoice.template_choice))
        logger.info("Selected template number: %s", template_number)
        selected_template = self.template_discovery.get_template_path(
            template_number, is_receipt
        )
        if not selected_template:
            logger.error("Template %s not found", invoice.template_choice)
            raise TemplateNotFound(f"Template {invoice.template_choice} not found.")
        template = self.env.get_template(os.path.basename(selected_template))
        logger.info("Rendering template: %s", selected_template)

        return template.render(context)

    def render_context(self, queryset: QuerySet) -> Dict[str, Any]:
        """Prepare the context data for rendering a quotation.

        Args:
            queryset (QuerySet): A queryset containing the invoice(s).

        Returns:
            Dict[str, Any]: The context data for rendering the quotation.
        """
        logger.info("Preparing context data for quotation")
        invoice = queryset
        total = invoice.total
        items = invoice.items.all()
        email = None
        phone = ""
        item_list = []
        custom_columns = set()
        additional_fields = invoice.notes if hasattr(invoice, "notes") else []
        contacts = invoice.contacts

        for contact in contacts:
            if "@" in contact and not email:
                email = contact
            elif contact.isdigit() and not phone:
                phone = contact

        for item in items:
            custom_data = Column.objects.filter(item=item).order_by("priority")
            custom_fields = {data.column_name: data.value for data in custom_data}
            custom_columns.update(custom_fields.keys())
            item_dict = {
                "description": item.description,
                "quantity": item.quantity,
                "measurement": item.measurement if item.measurement else "",
                "unit_price": item.unit_price,
                "total_price": item.total_price,
                "custom_data": custom_fields,
            }
            item_list.append(item_dict)

        context = {
            "title": invoice.title,
            "tracking_code": invoice.tracking_code,
            "items": item_list,
            "subtotal": total.subtotal,
            "tax_percentage": total.tax_percentage,
            "tax_amount": total.tax_amount,
            "discount_percentage": total.discount_percentage,
            "discount_amount": total.discount_amount,
            "concession_percentage": total.concession_percentage,
            "concession_amount": total.concession_amount,
            "grand_total": total.total_amount,
            "invoice_date": invoice.invoice_date,
            "customer_name": invoice.customer_name,
            "customer_email": email,
            "customer_phone": phone,
            "due_date": invoice.due_date,
            "status": invoice.status,
            "currency": invoice.currency,
            "logo_url": invoice.logo.url if invoice.logo else None,
            "sign_url": invoice.signature.url if invoice.signature else None,
            "stamp_url": invoice.stamp.url if invoice.stamp else None,
            "custom_columns": custom_columns,
            "additional_fields": additional_fields,
        }

        logger.info("Context prepared for invoice: %s", invoice.title)
        return context



from rest_framework import serializers

from sage_invoice.api.serializers import (
    CategorySerializer,
    ColumnSerializer,
    ExpenseSerializer,
    ItemSerializer,
)
from sage_invoice.models import Category, Invoice


class ContactFieldSerializer(serializers.Serializer):
    """
    Serializer to validate the structure of the contacts field (either phone or
    email).
    """

    phone = serializers.RegexField(
        regex=r"^\d{10,15}$",
        required=False,
        help_text="Phone number should only contain digits and be 10-15 characters long.",
    )
    email = serializers.EmailField(required=False, help_text="Valid email format.")

    def validate(self, data):
        """Ensure at least one of 'phone' or 'email' is provided."""
        if not data.get("phone") and not data.get("email"):
            raise serializers.ValidationError("Either phone or email must be provided.")
        return data


class NoteFieldSerializer(serializers.Serializer):
    """Serializer to validate the structure of the notes field (additional fields)."""

    label = serializers.CharField()
    content = serializers.CharField()


class InvoiceSerializer(serializers.HyperlinkedModelSerializer):
    contacts = ContactFieldSerializer(required=False)
    notes = serializers.ListField(
        child=serializers.DictField(child=serializers.CharField()), required=False
    )
    category = CategorySerializer(read_only=True)
    items = ItemSerializer(many=True, read_only=True)
    columns = ColumnSerializer(many=True, read_only=True)
    expense = ExpenseSerializer(read_only=True)

    class Meta:
        model = Invoice
        fields = [
            "url",
            "slug",
            "invoice_date",
            "customer_name",
            "tracking_code",
            "contacts",
            "status",
            "receipt",
            "notes",
            "currency",
            "due_date",
            "template_choice",
            "logo",
            "signature",
            "stamp",
            "category",
            "items",
            "columns",
            "expense",
        ]
        extra_kwargs = {
            "url": {"lookup_field": "slug"},
            "categories": {"lookup_field": "slug"},
            "items": {"lookup_field": "id"},
            "columns": {"lookup_field": "id"},
            "expense": {"lookup_field": "id"},
        }
        read_only_fields = ["slug"]



from rest_framework import serializers

from sage_invoice.models import Column


class ColumnSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Column
        fields = "__all__"
        lookup_field = "id"
        extra_kwargs = {
            "url": {"lookup_field": "id"},
            "invoice": {"lookup_field": "slug"},
            "item": {"lookup_field": "id"},
        }



from django.urls import path

from sage_invoice.views.invoice import (
    DownloadInvoicesView,
    GenerateInvoicesView,
    InvoiceDetailView,
    TemplateChoiceView,
)

urlpatterns = [
    path(
        "invoice/<slug:slug>/",
        InvoiceDetailView.as_view(),
        name="invoice_detail",
    ),
    path(
        "template-choices/<str:check>",
        TemplateChoiceView.as_view(),
        name="template_choices",
    ),
    path("generate-pdfs/", GenerateInvoicesView.as_view(), name="generate_pdfs"),
    path(
        "download-invoices/", DownloadInvoicesView.as_view(), name="download_invoices"
    ),
]
